# Copyright 2023 Camptocamp SA
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl)
{
    "name": "Test",
    "version": "1.0.0",
    "category": "Test Module",
    "author": "Camptocamp, Odoo Community Association (OCA)",
    "website": "https://github.com/OCA/oca-port",
    "license": "AGPL-3",
    "depends": ["base"],
    "data": [],
    "demo": [],
    "installable": True,
}
